/************************************************************
*  tpro.c - Device driver for TPRO-VME, Synchronizable Time
*           Generator with VME Interface.
*
*  Author: Tony Denault
*  NASA IRTF
*  2680 Woodlawn Drive, Honolulu, HI, 96822
*************************************************************/

#include <conf.h>
#include <kernel.h>
#include <file.h>
#include <ioctl.h>
#include <dldd.h>
#include <sysdev.h>
#include <types.h>

#include "tpro.h"
#include "tpro_prv.h"

extern char *sysbrk();
extern int  sysfree();

/************************************************************
             Device Driver ISR
************************************************************/

/*----------------------------------------------------------
**  Device Interrupt Service routine.
**  TPRO is set to clear the fifo every time the BIM is accessed.
**  Instead of reading the FIFO, read the D32 registers for
**  the time.
**----------------------------------------------------------
*/
static tpro_isr( s )
   register struct stuff *s;
{
   register long h, l;
   register long * lp;
   int   ps;
   short *sp;


   /* read Low & High A32 register for time */
   disable(ps);
   lp = (long*)(s->a32adr+REG_LWTIME);
   l  = *lp;

   lp = (long*)(s->a32adr + REG_HWTIME);
   h  = *lp;
   restore(ps);

   /* extract seconds & usec */
   s->usec = l&0x0f;
   s->usec += ((l>>4)&0x0f)*10;
   s->usec += ((l>>8)&0x0f)*100;
   s->usec += ((l>>12)&0x0f)*1000;
   s->usec += ((l>>16)&0x0f)*10000;
   s->usec += ((l>>20)&0x0f)*100000;
 
   s->sec = ((l>>24)&0x0f);
   s->sec += ((l>>28)&0x07)*10;

   /* extract min, hr, day */
   s->min = h&0x0f;
   s->min += ((h>>4)&0x07)*10;
 
   s->hour = ((h>>8)&0x0f);
   s->hour += ((h>>12)&0x03)*10;
 
   s->yday = ((h>>16)&0x0f);
   s->yday += ((h>>20)&0x0f)*10;
   s->yday += ((h>>24)&0x03)*100;

   /* update sem_counter & signal processes waiting on semaphores  */
   if( ++s->sem_counter >= 100 )
      s->sem_counter = 0;

   if( s->sem_counter % 2 )         /* 50 hz: 1, 3, 5, ... */
      sreset( &s->sem_50hz );

   if( (s->sem_counter % 10)==0 )   /* 10 hz, phase A: 0, 10, 20, ... */
      sreset( &s->sem_10hz_a );

   if( (s->sem_counter % 10)==3 )   /* 10 hz, phase B: 3, 13, 23, ... */
      sreset( &s->sem_10hz_b );

   if( (s->sem_counter % 10)==5 )   /* 10 hz, phase C: 5, 15, 25, ... */
      sreset( &s->sem_10hz_c );

   if( (s->sem_counter % 10)==8 )   /* 10 hz, phase D: 8, 18, 28, ... */
      sreset( &s->sem_10hz_d );

   if( (s->sem_counter % 20)==6 )   /* 5 hz:  6, 26, 36, ... */
      sreset( &s->sem_5hz );

   /* re-enable interrupts */
   sp = (short*)(s->a16adr + REG_BIMCNT3);
   *sp = BIMCNT3_DATA;

   return 0;
}

/************************************************************
             Device Driver Interface
************************************************************/

/*----------------------------------------------------------
** Open the device driver
**----------------------------------------------------------
*/
int tproopen( s, d, f)
   struct stuff *s;            /* driver data   */
   int d;                      /* device number */
   struct file * f;            /* file descriptor */
{
   return (OK);
}

/*----------------------------------------------------------
**  Close the device driver
**----------------------------------------------------------
*/
int tproclose()
{
   return (OK);
}

/*----------------------------------------------------------
**  Read from the device
**----------------------------------------------------------
*/
int tproread( s, f, buff, count)
   struct stuff *s;
   struct file *f;
   char *buff;
   register int count;
{
   return 0;
}

/*----------------------------------------------------------
**  Write to the device driver
**----------------------------------------------------------
*/
int tprowrite( s, f, buff, count)
   struct stuff *s;
   struct file *f;
   char *buff;
   register int count;
{
   return (count);
}

/*----------------------------------------------------------
**  Device control
**----------------------------------------------------------
*/
int tproioctl( s, f, command, tprop )
   struct stuff *s;
   struct file *f;
   int command;
   struct tproparms_t * tprop;
{
   short *sp;
   struct stuff *stuffp;

   /* process user's request */
   switch( command )
   {
      case TPRO_GETTIME:
         TPRO_GetTime( s, tprop );
         tprop->ret = 0;
         break;

      case TPRO_SETTIME:
         /* use semephore so only 1 task at a time can set the time     */
         /* Note TPRO_SetTime() include many sleep when talking to TPRO */
         swait( &s->iosem, -1);  
         TPRO_SetTime( s, tprop );
         tprop->ret = 0;
         ssignal( &s->iosem );
         break;

      case TPRO_TEST:
         tprop->ret = 0;
         break;

      case TPRO_INTER_EVENT:
         /* write to the internal event register */
         sp = (short*)(s->a16adr + REG_INTER_EVENT);
         *sp = 0;
         tprop->ret = 0;
         break;

      case TPRO_GETSTUFF:
         stuffp = (struct stuff*) tprop;
         stuffp->iosem = s->iosem;
         stuffp->a16adr = s->a16adr;
         stuffp->a32adr = s->a32adr;
         stuffp->yday  = s->yday;
         stuffp->hour  = s->hour;
         stuffp->min   = s->min;
         stuffp->sec   = s->sec;
         stuffp->usec  = s->usec;
         break;

      case TPRO_WAIT_ON_50HZ:
         /* Sleeps until the ISR waits up the process (50Hz Please). */
         swait( &s->sem_50hz, 0);  
         break;

      case TPRO_WAIT_ON_10HZ_A:
         /* Sleeps until the ISR waits up the process (10Hz Please). */
         swait( &s->sem_10hz_a, 0);  
         break;

      case TPRO_WAIT_ON_10HZ_B:
         /* Sleeps until the ISR waits up the process (10Hz Please). */
         swait( &s->sem_10hz_b, 0);  
         break;

      case TPRO_WAIT_ON_10HZ_C:
         /* Sleeps until the ISR waits up the process (10Hz Please). */
         swait( &s->sem_10hz_c, 0);  
         break;

      case TPRO_WAIT_ON_10HZ_D:
         /* Sleeps until the ISR waits up the process (10Hz Please). */
         swait( &s->sem_10hz_d, 0);  
         break;

      case TPRO_WAIT_ON_5HZ:
         /* Sleeps until the ISR waits up the process ( 5Hz Please). */
         swait( &s->sem_5hz, 0);  
         break;

      case TPRO_SYNC_SEM:
         TPRO_SyncSem( s );
         break;

      default:
         tprop->ret = -1;
         break;
   }

   return (OK);
}

/*----------------------------------------------------------
**  Selection
**----------------------------------------------------------
*/
int tproselect()
{
   return (OK);
}

/*----------------------------------------------------------
**  Initial device installation. Allocates memory for
**  tproinfo.
**----------------------------------------------------------
*/
struct stuff * tproinstall( info )
   struct tproinfo_t * info;
{
   struct stuff *s;
   register short *sp;

   s = (struct stuff *)sysbrk((long)sizeof(struct stuff));
   s->a32adr = info->a32adr;
   s->a16adr = info->a16adr;
   s->iosem = 1;
   s->sem_counter = 0;
   s->sem_50hz = 0;
   s->sem_10hz_a = 0;
   s->sem_10hz_b = 0;
   s->sem_10hz_c = 0;
   s->sem_10hz_d = 0;
   s->sem_5hz = 0;
   s->yday = 0;
   s->hour = 0;
   s->min  = 0;
   s->sec  = 0;
   s->usec = 0;

   /* Establish ISR routine & enable interrupts */
   iointset( BIMVEC3_DATA*4, tpro_isr, (char*)s);

   /* set vector number on hardware */
   sp = (short*)(s->a16adr + REG_BIMVEC3);
   *sp = BIMVEC3_DATA;

   /* enable interrupts */
   sp = (short*)(s->a16adr + REG_BIMCNT3);
   *sp = BIMCNT3_DATA;
   return s;
}

/*----------------------------------------------------------
**  Device un-installation
**----------------------------------------------------------
*/
int tprouninstall( s )
   struct stuff *s;
{
   iointclr( BIMVEC3_DATA*4 );
   sysfree( s, (long)sizeof(struct stuff));
   return (OK);
}


/*--------------------------------------------------------------
**  TPRO_csleep() - a sleep function in units of 1/100 seconds.
**                csec = 1 means sleep for 0.01 seconds.
**--------------------------------------------------------------
*/
#if ANSI
int TPRO_csleep( int csec )
#else
int TPRO_csleep( csec )
   int csec;
#endif
{
   int *wakeup;
 
   wakeup = (int *) sysbrk(sizeof(int));
   *wakeup = 0;
   timeout(TPRO_csleeph, wakeup, csec);
   swait(wakeup, 0);
   sysfree(wakeup);
}

#if ANSI
int TPRO_csleeph( int *s )
#else
int TPRO_csleeph( s )
   int *s;
#endif
{
   ssignal(s);
}

/************************************************************
*                Hardware specific functions                *
*************************************************************/

/*-------------------------------------------------------------
** TPRO_GetStatus() - Return the contents of the status register.
**------------------------------------------------------------
*/
int TPRO_GetStatus( s )
   register struct stuff *s;            /* driver data   */
{
   register short *sp;

   sp = (short*)(s->a16adr + REG_STATUS);
   return *sp;
}

/*-------------------------------------------------------------
** TPRO_ReadFIFO() - reads the next byte from the FIFO register.
**------------------------------------------------------------
*/
int TPRO_ReadFIFO( s )
   register struct stuff *s;            /* driver data   */
{
   register int cnt;
   register short *sp;

   /* Wait until fifo is not empty */
   cnt = 0;
   while( TPRO_GetStatus( s ) & STATUS_FIFO_EMPTY )  /* output ready when low */
   {
      if( cnt++ > 10000)
      {
         /* cprintf("TPRO_ReadFIFO() Timeout\n"); */
         return 0;
      }
   }

   /* read fifo and return character */
   sp = (short*)(s->a16adr + REG_FIFO_READ);
   cnt = *sp & 0x00ff;
   /* cprintf("TPRO_ReadFIFO() [0x%x]\n", cnt); */
   return cnt;
}

/*-------------------------------------------------------------
** TPRO_GetTime() - Reads the time from the TPRO board.
**------------------------------------------------------------
*/
#if ANSI
int TPRO_GetTime( struct stuff *s, struct tproparms_t * t )
#else
int TPRO_GetTime( s, t ) 
   struct stuff *s;
   register struct tproparms_t * t;
#endif
{
   register long l, h;
   int  ps;
   long * lp;
   short * sp;

   /* read Low & High A32 register for time */
   disable(ps);
   lp = (long*)(s->a32adr+REG_LWTIME);
   l  = *lp;

   lp = (long*)(s->a32adr + REG_HWTIME);
   h  = *lp;
   restore(ps);

	/* extract seconds & usec */
   t->usec = l&0x0f;
   t->usec += ((l>>4)&0x0f)*10;
   t->usec += ((l>>8)&0x0f)*100;
   t->usec += ((l>>12)&0x0f)*1000;
   t->usec += ((l>>16)&0x0f)*10000;
   t->usec += ((l>>20)&0x0f)*100000;
 
   t->sec = ((l>>24)&0x0f);
   t->sec += ((l>>28)&0x07)*10;

   /* extract min, hr, day */
   t->min = h&0x0f;
   t->min += ((h>>4)&0x07)*10;

   t->hour = ((h>>8)&0x0f);
   t->hour += ((h>>12)&0x03)*10;

   t->yday = ((h>>16)&0x0f);
   t->yday += ((h>>20)&0x0f)*10;
   t->yday += ((h>>24)&0x03)*100;

   return 0;
}
 
/*-------------------------------------------------------------
** TPRO_SetTime() - Sets the time on the TPRO board.
**------------------------------------------------------------
*/
#if ANSI
int TPRO_SetTime( struct stuff *s, struct tproparms_t * t )
#else
int TPRO_SetTime( s, t )
   struct stuff *s;
   register struct tproparms_t * t;
#endif
{
   register short *sp;

   sp = (short*)(s->a16adr+REG_COMMAND);

   *sp = 0x00f0;      /* ZERO time set buffer command */
	TPRO_csleep( 1 );

   *sp = ((t->yday / 100) & 0x0f) | 0x0050;
	TPRO_csleep( 1 );

   *sp = (((t->yday/10) % 10) & 0x0f) | 0x0060;
	TPRO_csleep( 1 );

   *sp = ((t->yday % 10) & 0x0f) | 0x0070;
	TPRO_csleep( 1 );

   *sp = ((t->hour/10) & 0x0f) | 0x0080;
	TPRO_csleep( 1 );

   *sp = ((t->hour % 10) & 0x0f) | 0x0090;
	TPRO_csleep( 1 );

   *sp = ((t->min/10) & 0x0f) | 0x00a0;
	TPRO_csleep( 1 );

   *sp = ((t->min % 10) & 0x0f) | 0x00b0;
	TPRO_csleep( 1 );

   *sp = ((t->sec/10) & 0x0f) | 0x00c0;
	TPRO_csleep( 1 );

   *sp = ((t->sec % 10) & 0x0f) | 0x00d0;
	TPRO_csleep( 1 );

   *sp = 0x00e0;    /* GO command */

   return 0;
}

/*-------------------------------------------------------------
** TPRO_SyncSem() - Tries to sync the semephore counter with
**   the internal clock so wait on pulses occur on at nice time boundaries.
**------------------------------------------------------------
*/
#if ANSI
int TPRO_SyncSem( struct stuff *s )
#else
int TPRO_SyncSem( s ) 
   struct stuff *s;
#endif
{
   int   ps;
   register long l, usec;
   long h;
   long * lp;

   /* read Low & High A32 register for time */
   disable(ps);
   lp = (long*)(s->a32adr+REG_LWTIME);
   l  = *lp;

   lp = (long*)(s->a32adr + REG_HWTIME);
   h  = *lp;
   restore(ps);

	/* extract seconds & usec */
   usec = l&0x0f;
   usec += ((l>>4)&0x0f)*10;
   usec += ((l>>8)&0x0f)*100;
   usec += ((l>>12)&0x0f)*1000;
   usec += ((l>>16)&0x0f)*10000;
   usec += ((l>>20)&0x0f)*100000;
 
   s->sem_counter = usec  / 10000;
   /*
   kprintf(" %06ld %02ld \n", usec, s->sem_counter );
   */
}

/*---------------------------- eof ------------------------------*/

